import uuid


class ONone:
    def __getattribute__(self, *args, **kwargs):
        return ONone()

    def __getitem__(self, key):
        return ONone()

    def __str__(self):
        return ''

    def __repr__(self):
        return ''

    def __bool__(self):
        return False


class OInt(int):
    def __getattribute__(self, key):
        try:
            value = super(OInt, self).__getattribute__(key)
        except AttributeError as ae:
            value = ONone()
        return value


class OFloat(float):
    def __getattribute__(self, key):
        try:
            value = super(OFloat, self).__getattribute__(key)
        except AttributeError as ae:
            value = ONone()
        return value


class OString(str):
    def __getattribute__(self, key):
        try:
            value = super(OString, self).__getattribute__(key)
        except AttributeError as ae:
            value = ONone()
        return value


class ODict(dict):
    __helper__ = {}

    def __init__(self, d: dict = {}, **kwargs):
        unique_key = str(uuid.uuid4())
        self.__helper__[unique_key] = {}
        self.__unique_key__ = unique_key
        for key, value in d.items():
            setattr(self, key, value)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __to_dict__(self):
        me = self.__dict__
        h = self.__helper__[self.__unique_key__]
        sum = {}
        sum.update(me)
        sum.update(h)
        del sum['__unique_key__']
        return sum

    def __add__(self, val):
        r_val = None
        try:
            r_val = self.__to_dict__() + val
        except:
            r_val = val
        return r_val

    def __sub__(self, val):
        r_val = None
        try:
            r_val = self.__to_dict__() - val
        except:
            r_val = -val
        return r_val

    def __str__(self):
        return str(self.__to_dict__())

    def __repr__(self):
        return str(self.__to_dict__())

    def __getattribute__(self, key):
        try:
            value = super(ODict, self).__getattribute__(key)
        except AttributeError as ae:
            try:
                value = self.__helper__[self.__unique_key__][key]
            except:
                value = ODict()
                self.__setattr__(key, value)
        return value

    def __bool__(self):
        if len(self.__to_dict__()) == 0:
            return False
        return True

    def __setattr__(self, key, value):
        number = False
        try:
            r = float(key)
            number = True
        except:
            pass

        if number:
            self.__helper__[self.__unique_key__][key] = value
            return

        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        super(ODict, self).__setattr__(key, value)

    def keys(self):
        return self.__to_dict__().keys()

    def update(self, d: dict):
        for key, value in d.items():
            self.__setattr__(key, value)

    def dict(self):
        from ast import literal_eval
        mystr = str(self)
        this_dict = literal_eval(mystr)
        return this_dict


class OList(list):
    def __getitem__(self, key):
        try:
            value = super(OList, self).__getitem__(key)
        except IndexError as ie:
            value = ONone()
        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        return value

    def __getattribute__(self, key):
        try:
            value = super(OList, self).__getattribute__(key)
        except AttributeError as ae:
            value = ONone()
        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        return value

    def append(self, value):
        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        super(OList, self).append(value)

    def insert(self, index, value):
        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        super(OList, self).insert(index, value)

    def __setitem__(self, key, value):
        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        super(OList, self).__setitem__(key, value)

    def __setattr__(self, key, value):
        if type(value) is int:
            value = OInt(value)
        elif type(value) is float:
            value = OFloat(value)
        elif type(value) is str:
            value = OString(value)
        elif type(value) is list:
            value = OList(value)
        elif type(value) is dict:
            value = ODict(value)
        super(OList, self).__setattr__(key, value)



import uuid
import json
from inspect import currentframe


def printwln(*args, **kwargs):
    """Print with the origin line number"""
    cf = currentframe()
    line = cf.f_back.f_lineno
    print(f"ln {line}: ", *args, **kwargs)


def print_line_center(text: str) -> str:
    print()
    print(text)
    print()
    return f'\n{text}\n'


def get_unique_key():
    return str(uuid.uuid4())


def print_prev(*args, **kwargs):
    CURSOR_UP_ONE = '\x1b[1A'
    ERASE_LINE = '\x1b[2K'
    print(CURSOR_UP_ONE, end='')
    print(ERASE_LINE, end='')
    print(*args, **kwargs)


def print_json(object: dict):
    # with datetime objects
    json_object = json.dumps(object, indent=4, default=str)
    print(json_object)


def valida_dato(dato: any):
    if str(dato).lower == 'nan' or str(dato).lower == 'none' or str(dato).lower == 'undefined' or str(dato).lower == 'null' or str(dato).lower == '':
        return None
    else:
        return dato


def get_d(d: dict, key: str, none=False, to_parse=None, *args, **kwargs) -> any:
    """Validate field in dict and return it or a certain value"""
    if key in d and d[key] is not None:
        if to_parse:
            try:
                return to_parse(d[key])
            except:
                return d[key]
        return d[key]
    elif 'default' in kwargs:
        return kwargs['default']
    elif none:
        return None
    else:
        raise Exception(f'Error: "{key}" not found')


def get_separated_number(n: any) -> str:
    """Return a separated number string"""
    n = str(n).replace(' ', '')
    try:
        n = float(n)
    except:
        raise Exception('Cannot convert to number')
    return f'{n:,.2f}'


def get_currency(n: any) -> str:
    """Return a currency string"""
    n = str(n).replace(' ', '')
    try:
        n = float(n)
    except:
        raise Exception('Cannot convert to number')

    n_final = round(n, 2)
    n_final = ('-' if n_final < 0 else '') + \
        f'$ {get_separated_number(abs(n_final))}'
    return n_final


def destructure_d(*args, **kwargs):
    """get value from names varialbes if exist in dictionary else get None
    Args: 
        1.- position de dict to get values
        2.- a list with variables names or values separated with ,
    
    Examples:
        d_1 = {'a':1,'b':2,'c':3,'d':4,'e':5}
        a,b,g,c = destructure_d(d_1, 'a','b','g','c')
        a >> 1
        b >> 2
        g >> None
        c >> 3

        a,e,d1,d2 = destructure_d(d_1, ['a','e','d1','d2'])
        a >> 1
        e >> 5
        d1 >> None
        d2 >> None
    """
    args = list(args)
    the_dict = args.pop(0)
    
    values = []
    if type(args[0]) == list:
        for key in args[0]:
            values.append(get_d(the_dict, key, default=None))
    else:
        for key in args:
            values.append(get_d(the_dict, key, default=None))
    return values









class TextFormat:
    def without_acute(self, text: str) -> str:
        acutes = {
            'á': 'a',
            'é': 'e',
            'í': 'i',
            'ó': 'o',
            'ú': 'u',
            'Á': 'A',
            'É': 'E',
            'Í': 'I',
            'Ó': 'O',
            'Ú': 'U'
        }
        for acute in acutes:
            text = text.replace(acute, acutes[acute])
        return text

    def normalize_spaces(self, text: str) -> str:
        special_spaces = {
            '{ ': '{',
            ' {': '{',
            '} ': '}',
            ' }': '}',
            '[ ': '[',
            ' [': '[',
            '] ': ']',
            ' ]': ']',
            '( ': '(',
            ' (': '(',
            ') ': ')',
            ' )': ')',
        }
        if not '  ' in text:
            return text
        while True:
            text = text.replace('  ', ' ')
            if not '  ' in text:
                break
        for special_space in special_spaces:
            text = text.replace(special_space, special_spaces[special_space])
        return text

    def unique_string_no_space(self, items: list) -> str:
        if type(items) is str:
            items = items.split()
        elif type(items) is not list:
            items = [items]
        text = ''
        for item in items:
            if not item:
                continue
            item = str(item)
            item = self.without_acute(item)
            text += item.lower().replace(' ', '')
        return text

    def unique_string(self, items: str) -> str:
        text = ''
        for item in items:
            if not item:
                continue
            item = self.without_acute(item)
            item = self.normalize_spaces(item)
            text += item.lower() + ' '
        text = text[:-1]
        return text

    def normal_text(self, items: list) -> str:
        if type(items) is str:
            items = items.split()
        elif type(items) is not list:
            items = [items]
        text = ''
        for item in items:
            if not item:
                continue
            item = self.normalize_spaces(item)
            text += item + ' '
        text = text[:-1]
        text = self.normalize_spaces(text)
        return text

    def cambiar_estado(self, estado):
        ESTADOS = {
            'bajacalifornia': 'BCN',
            'bajacalifornianorte': "BCN",
            'bajacaliforniasur': "BCS",
            'aguascalientes': "AGS",
            'campeche': "CAM",
            'chihuahua': "CHI",
            'chiapas': "CHS",
            'coahuila': "COH",
            'colima': "COL",
            'distritofederal': "CDMX",
            'df': "CDMX",
            'ciudaddemexico': "CDMX",
            'cdmx': "CDMX",
            'durango': "DUR",
            'guerrero': "GRO",
            'guanajuato': "GTO",
            'hidalgo': "HGO",
            'jalisco': "JAL",
            'edo.demexico': "MEX",
            'estadodemexico': "MEX",
            'estadomexico': "MEX",
            'edomexico': "MEX",
            'mexico': "MEX",
            'michoacan': "MIC",
            'morelos': "MOR",
            'nayarit': "NAY",
            'nuevoleon': "NL",
            'oaxaca': "OAX",
            'puebla': "PUE",
            'quintanaroo': "QR",
            'queretaro': "QRO",
            'sinaloa': "SIN",
            'sanluispotosi': "SLP",
            'sonora': "SON",
            'tabasco': "TAB",
            'tamaulipas': "TAM",
            'tlaxcala': "TLX",
            'veracruz': "VER",
            'yucatan': "YUC",
            'zacatecas': "ZAC",
        }
        unique_edo = self.unique_string_no_space(estado.split())
        if unique_edo in ESTADOS:
            estado = ESTADOS[unique_edo]
        return estado



import os
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import smtplib


class ErrorEmail:
    def __init__(self, message, email_settings, name_project='No hay nombre del projecto'):
        self.msg = MIMEMultipart()
        self.server = smtplib.SMTP_SSL(email_settings['smtp_server'])
        self.server.login(email_settings['user'], email_settings['password'])
        self.message = message
        self.sender = email_settings['sender']
        self.subject = f'ERROR EN {name_project}'
        self.receiver = email_settings['receiver']

    def send(self):
        self.msg['From'] = self.sender
        self.msg['Subject'] = self.subject
        self.msg['To'] = self.receiver
        self.msg.attach(MIMEText(self.message, 'plain'))
        self.server.sendmail(
            self.msg['From'], self.msg['To'], self.msg.as_string())
        self.server.quit()


class CatchErrors:
    def __init__(self, name_project: str = 'No hay nombre del projecto', email_settings: dict = None, root_path: str = None):
        self.name_project = name_project
        self.email_settings = email_settings
        self.root_path = str(root_path)
        if not self.email_settings:
            self.email_available = False
        else:
            self.email_available = True

    def show_error(self, e: Exception, send_email: bool = False, root_path: str = None, extra: str = '') -> str:
        import os
        import datetime
        if root_path:
            self.root_path = str(root_path)

        com_path = '' if not self.root_path else self.root_path
        if com_path.endswith('/'):
            com_path = com_path[:-1]

        now = datetime.datetime.now()
        now = now.strftime("%d/%m/%Y %H:%M:%S")
        info_exc = os.sys.exc_info()
        et, eo, et = info_exc
        errs = []
        errs.append(['file', 'line', 'function', 'code'])
        paths = []
        codes = []
        max_len = 0
        max_file = 0
        max_func = 0
        while (et):
            file = et.tb_frame.f_code.co_filename
            line = et.tb_lineno
            function_data = str(et.tb_frame).split()
            function_data = function_data[-1][:-1]
            code = ''
            with open(file, 'r') as f:
                Lines = f.readlines()
                code_text = Lines[line - 1].replace('\n', '')
                try:
                    ant = Lines[line - 2].replace('\n', '')
                except:
                    ant = ''
                try:
                    aft = Lines[line].replace('\n', '')
                except:
                    aft = ''
                codes.append(
                    f'{line - 1}: {ant}\n{line}: {code_text}\n{line + 1}: {aft}')

            # Validate root path
            p = file.split('/')[:-1]
            ac_path = '/'.join(p)
            if not self.root_path:
                if not com_path:
                    com_path = ac_path

                if ac_path != com_path and ac_path not in paths:
                    paths.append(ac_path)
                    ant = com_path.split('/')
                    new = []
                    for a, b in zip(p, ant):
                        if a != b:
                            break
                        else:
                            new.append(a)
                    com_path = '/'.join(new)

            code_text = code_text.replace('\n', '')
            max_len = max(len(code_text), max_len)
            max_file = max(len(file), max_file)
            max_func = max(len(function_data), max_func)
            errs.append([f'{file}:{line}', line, function_data, code_text])
            et = et.tb_next

        traceback = ''
        tb = ''
        max_file = max_file - len(com_path)

        i = 0
        for row in errs:
            p = row.pop(0)
            p = p.replace(com_path + '/', '')
            if i > 0:
                row[len(row) - 1] = codes[i-1]
                tb += 'file: {}\tline: {}\tfunc: {}\tcode: \n{}\n\n'.format(
                    p, *row)
            else:
                tb += '\n'
            i += 1

        extra = f'Extra: {extra}\n' if extra else ''
        error = f'ERROR INFO\nTipo: {et}\nCommon Path: {com_path}\n{tb}\nError: {e}\n{extra}Fecha: {now}'
        if send_email:
            if not self.email_available:
                return 'No hay configuración de email'
            email = ErrorEmail(error, self.email_settings, self.name_project)
            try:
                email.send()
            except:
                return 'Revise los datos del email'

        return error
